/* pmbt_main.c
   Enumeration of polyominoes by backtracking.

   Ambros Marzetta

   size	solutions	time [s] (DEC 3000-M700 Alpha)
   2	1
   3	2
   4	5
   5	12
   6	35
   7	108
   8	369
   9	1285
   10	4655		2
   11	17073		11
   12	63600		44
   13	238591		196
   
 * $Source: /home/schilte/marzetta/work/distr/zram/RCS/pmbt_main.c,v $
 * $Revision: 1.1 $
 * $Date: 1998/06/17 18:17:42 $
 * $Log: pmbt_main.c,v $
 * Revision 1.1  1998/06/17  18:17:42  marzetta
 * Initial revision
 *
 */

#include <stdlib.h>		/* exit() */
#include <stdio.h>
#include "z_io.h"		/* getopt() */
#include "z_time.h"
#include "z_enumps.h"
#include "z_signal.h"
#include "pm_both.h"
#include "z_bt.h"
#include "z_par.h"

static z_bt_ProcSetT pm_procset1;
static int pm_procsetnum1, pm_GlobalTag, pm_NodeTag;

enum outputType {ot_normal, characterGraphic, postScript, noOutput};

enum pm_state {state_empty, state_used, state_near, state_far};

#if 0
const char StateChar[] = ".#nf";
#else
const char StateChar[] = ".#..";
#endif

typedef struct pm_Global {
  Z_NODE_HEADER;
  int pm_n, pm_h, pm_w, pm_numPt;
  enum outputType outputType;
} pm_Global, *pm_GlobalP;

typedef struct pm_Node {
  Z_NODE_HEADER;
  int numUsed;
  enum pm_state pt[1];
} pm_Node, *pm_NodeP;

#ifdef SEQUENTIAL
static long count;
#endif /* SEQUENTIAL */

static void PrintFull(pm_GlobalP global, pm_NodeP node)
{
  /* Print the whole matrix - use this for debugging */
  int i, j;
  for (j=0; j<global->pm_h; j++) {
    for (i=0; i<global->pm_w; i++)
      printf("%c", StateChar[node->pt[j*global->pm_w+i]]);
    printf("\n");
  }
  printf("\n");
} /* PrintFull */

static boolean LexMin(pm_GlobalP global, pm_NodeP node)
{
  /* Return true if node is lexicographically minimum among its
     symmetric polyominoes. */
  int x[MAXN], y[MAXN], a[MAXN];
  int i, count, xmin, xmax, ymin, ymax, xx, yy;

  count = 0;
  xmin = ymin = MAXN * 3;
  xmax = ymax = -1;
  for (i=global->pm_numPt-1; i>=0; --i) /* find min/max coordinates */
    if (node->pt[i] == state_used) {
      xx = i % global->pm_w;
      yy = i / global->pm_w;
      if (xx < xmin) xmin = xx;
      if (xx > xmax) xmax = xx;
      if (yy < ymin) ymin = yy;
      if (yy > ymax) ymax = yy;
      x[count] = xx;
      y[count] = yy;
      count ++;
    }
  assert(ymax == global->pm_h-2);
  assert(count == global->pm_n);
  for (i=count-1; i>=0; --i) {
    x[i] -= xmin;		/* translate */
    y[i] -= ymin;
    a[i] = x[i] * MAXN + y[i];	/* store points in array a */
  }
  xmax -= xmin;
  ymax -= ymin;
  return lexmax(a, count, xmax, ymax);
} /* LexMin */

#define MakeNear(old, index) {old = node->pt[index]; \
			      if (old == state_far) \
				node->pt[index] = state_near; \
			     }

boolean pm_BranchInPlace(pm_GlobalP global, pm_NodeP node,
			 z_bt_DoProc whatToDo, void *ref)
{
  int i;
  enum pm_state left, right, top, bottom;
  
  if (node->numUsed == global->pm_n) {
    if (LexMin(global, node)) {
#ifdef SEQUENTIAL
      count ++;
#endif /* SEQUENTIAL */
      switch (global->outputType) {
      case characterGraphic:
      case postScript:
	z_Save((z_NodeP)global, (z_NodeP)node);
	break;
      case noOutput:
	break;
      default:
	PrintFull(global, node);
      } /* switch */
    } /* LexMin */
  } else {
    i = global->pm_numPt-1;   /* find the first place adjacent to a used one */
    while (i && node->pt[i] != state_near) i--;
    if (node->pt[i] == state_near) {
      node->pt[i] = state_empty; /* first branch */
      whatToDo((z_NodeP) node, ref);
      node->pt[i] = state_used;	/* second branch */
      node->numUsed ++;
      MakeNear(left, i-1);
      MakeNear(right, i+1);
      MakeNear(top, i-global->pm_w);
      MakeNear(bottom, i+global->pm_w);
      whatToDo((z_NodeP) node, ref);
      node->pt[i] = state_near;	/* revert */
      node->numUsed --;
      if (left == state_far) node->pt[i-1] = state_far;
      if (right == state_far) node->pt[i+1] = state_far;
      if (top == state_far) node->pt[i-global->pm_w] = state_far;
      if (bottom == state_far) node->pt[i+global->pm_w] = state_far;
    }
  }
  return 0;
} /* pm_BranchInPlace */

/* Postscript output ------------------------------------------------------ */

void pm_PSHeader(pm_GlobalP global, FILE *psfile, float xsize, float ysize)
{
  float side;

  side = ysize / (global->pm_h - 2);
  fprintf(psfile, "/side %7.2f def\n", side);
  fprintf(psfile, "/s {\n");
  fprintf(psfile, "  newpath side mul exch side mul exch moveto\n");
  fprintf(psfile, "  side 0 rlineto 0 side rlineto\n");
  fprintf(psfile, "  0 side sub 0 rlineto closepath\n");
  fprintf(psfile, "  gsave 0.5 setgray fill grestore stroke} def\n");
  fprintf(psfile, "\n");
} /* pm_PSHeader */

void FindMin(pm_GlobalP global, pm_NodeP node, int *xmin, int *ymin)
{
  /* Find left and top part of bounding box.
     Called by pm_PSPicture and MakeLine */
  int pt, x, y;
  
  *xmin = *ymin = global->pm_w;
  for (pt=0; pt<global->pm_numPt; pt++) {
    if (node->pt[pt] == state_used) {
      x = pt % global->pm_w;
      y = pt / global->pm_w;
      if (x < *xmin) *xmin = x;
      if (y < *ymin) *ymin = y;
    } /* if pt */
  } /* for pt */
} /* FindMin */

void pm_PSPicture(pm_GlobalP global, pm_NodeP node, FILE *psfile,
		  float xsize, float ysize)
{
  int pt, xmin, ymin, x, y;
  
  assert(global && global->tag == pm_GlobalTag);
  FindMin(global, node, &xmin, &ymin);
  for (pt=0; pt<global->pm_h*global->pm_w; pt++) {
    if (node->pt[pt] == state_used) {
      x = pt % global->pm_w - xmin;
      y = pt / global->pm_w - ymin;
      fprintf(psfile, "%d %d s\n", x, y);
    } /* if pt */
  } /* for pt */
} /* pm_PSPicture */


/* Character graphic output ----------------------------------------------- */

#define CHARPERLINE 80

void MakeLine(pm_GlobalP global, pm_NodeP *n, int k)
{
  int i, j, line, pt;
  int xmin[CHARPERLINE / 3], ymin[CHARPERLINE / 3];
  boolean fill;
  
  for (i=0; i<k; i++) FindMin(global, n[i], &xmin[i], &ymin[i]);
  
  for (line=1; line<global->pm_h-1; line++) { /* line of text file */
    for (i=0; i<k; i++) {	/* element i */
      for (j=0; j<global->pm_n; j++) { /* character j */
	pt = line * global->pm_w + xmin[i] + j;
	fill = (pt < global->pm_numPt && n[i]->pt[pt] == state_used);
	printf("%c", fill ? '#' : ' ');
      }
      if (i < k-1) printf("  "); /* space between elements */
    }
    printf("\n");
  }
  printf("\n");			/* empty line between element rows */
  
  for (i=0; i<k; i++) z_FreeNode((z_NodeP)n[i]);
} /* MakeLine */

void MakeCharacterGraphic(pm_GlobalP global)
{
  int32 elem_count;
  int elemPerLine, i, k;
  pm_NodeP n[CHARPERLINE / 3];
  
  z_enumps_Rewind(&elem_count);
  elemPerLine = CHARPERLINE / (global->pm_n+2);
  k = 0;
  for (i=0; i<elem_count; i++) {
    n[k++] = (pm_NodeP) z_Load();
    if (k == elemPerLine) {
      MakeLine(global, n, k);
      k = 0;
    }
  }
  if (k) MakeLine(global, n, k);
} /* MakeCharacterGraphic */


/* Initialization -------------------------------------------------------- */

void initproc(void)
{
  z_bt1_Init();
  z_bt2_Init();
  z_enumps_init();
  z_signal_Init();

  pm_GlobalTag = z_InstallClass(pm_Global, 0, 0, 0);
  pm_NodeTag = z_InstallClass(pm_Node, 0, 0, 0);
  pm_procset1.bt_BranchInPlace = (z_bt_BranchProc)pm_BranchInPlace;
  pm_procsetnum1 = z_InstallProcSet((z_ProcSetP) &pm_procset1);
} /* initproc */


/* Main program ----------------------------------------------------------- */

static int Usage(const char *name)
{
  fprintf(stderr, "Usage: %s [-1chnprt] n\n\n", name);
  fprintf(stderr, "Options: -1  sequential\n");
  fprintf(stderr, "         -c  character graphic output\n");
  fprintf(stderr, "         -h  help (produce this message)\n");
  fprintf(stderr, "         -n  no output (implies time measurement)\n");
  fprintf(stderr, "         -p  PostScript output\n");
  /*fprintf(stderr, "         -r  restart\n");*/
  fprintf(stderr, "         -t  time measurement and count of polyominoes\n");
  return 4;
} /* Usage */

int mmain(int argc, char **argv, char **envp)
{
  pm_GlobalP global;
  pm_NodeP root;
  int i, j, n, w, h, numpt;
  double time;
  unsigned int index;
  enum outputType outputType = ot_normal;
  boolean showTime=0;
  boolean sequential=0;
  boolean restart=0;
  int procsetnum;
  char c;

  while ((c=getopt(argc, argv, "1chnprt")) != -1)
    switch(c) {
    case 'c':
      outputType = characterGraphic;
      break;
    case 'n':
      outputType = noOutput;
      showTime = 1;
      break;
/*    case 'r':
      restart = 1;
      break;*/
    case 'p':
      outputType = postScript;
      sequential = 1;
      break;
    case 't':
      showTime = 1;
      break;
    case '1':
      sequential = 1;
      break;
    default:
      return Usage(argv[0]);
    }
  
  if (argc != optind+1) return Usage(argv[0]);
  
  n = atoi(argv[optind]);
  if (n < 2) {
    Usage(argv[0]);
    fprintf(stderr, "n must be >= 2\n");
    return 3;
  }
  
  h = (n+1) / 2 + 2;
  w = 2 * n - 2 + 2;
  numpt = h*w;
  if (numpt > 30000) {
    fprintf(stderr, "%d is too large for n\n", n);
    exit(5);
  }
  global = (pm_GlobalP) z_NewNode(pm_GlobalTag, sizeof(pm_Global));
  if (!global) exit(1);
  global->pm_n = n;
  global->pm_h = h;
  global->pm_w = w;
  global->pm_numPt = h*w-n+2;
  global->outputType = outputType;
  if (! restart) {
    root = (pm_NodeP) z_NewNode(pm_NodeTag,
				sizeof(pm_Node) + numpt * sizeof(root->pt[0]));
    for (i=numpt-1; i>=0; --i) root->pt[i] = state_far;
    root->numUsed = 1;
    root->pt[(n+1)/2*w + n] = state_used;	/* initial tile */
    root->pt[(n+1)/2*w + n-1] = state_near;	/* left neighbour */
    root->pt[(n-1)/2*w + n] = state_near;	/* top neighbour */
    for (i=(n+1)/2*w+n+1; i<numpt; i++)	/* bottom two rows */
      root->pt[i] = state_empty;
    for (i=w-1; i>=0; --i) {
      root->pt[i] = state_empty; /* top row */
    }
    for (i=h-1; i>=0; --i) {
      root->pt[i*w] = state_empty; /* leftmost row */
      root->pt[i*w+w-1] = state_empty; /* rightmost row */
    }
  }
  
  switch (outputType) {
  case noOutput:
    procsetnum = pm_procsetnum1; /* no output */
    break;
  case characterGraphic:
  case postScript:
    procsetnum = pm_procsetnum1; /* save */
    z_enumps_Open("/tmp/polyomino.tmp");
    break;
  default:
    procsetnum = pm_procsetnum1; /* print */
    break;
  }
  
  time = z_Gtime();
  /*if (restart)*/
  /*z_bt_Restart(procsetnum, (z_bt_GlobalP) global, sequential);*/
  /*else*/
#if 1
  z_bt_EnumerateDFExp(procsetnum,
		      (z_bt_GlobalP) global, (z_NodeP)root, sequential);
#else
  z_bt_EnumerateDepthFirst(procsetnum,
		      (z_bt_GlobalP) global, (z_NodeP)root);
#endif
  time = z_Gtime() - time;

  if (outputType == characterGraphic)
    MakeCharacterGraphic(global);
  else if (outputType == postScript) {
    z_enumps_ratio = 0.5;
    z_enumps_frame = 0;
    z_enumps_borderFraction = 1/18.0;
    z_PrintPS((z_NodeP)global, stdout,
	      (z_enumps_HeaderProc)pm_PSHeader,
	      (z_enumps_PictureProc)pm_PSPicture);
  }


  if (showTime) {
#ifdef SEQUENTIAL
    printf("There are %ld polyominoes of order %d\n", count, n);
#endif /* SEQUENTIAL */
    if (sequential)
      printf("Time:%7.2f (sequential) n = %d\n",
	     time, n);
    else
      printf("Time:%7.2f (%2d processor%s n = %d\n",
	     time, z_numprocs, z_numprocs>1?"s)":") ", n);
  }
  return 0;
} /* mmain */
