/* vc_main.c

   Vertex cover problem main program

   Ambros Marzetta

   94-11-17
   For more information see our WWW page at
   
         http://wwwjn.inf.ethz.ch/ambros/zram.html
	 
 * $Source: /home/schilte/marzetta/work/distr/zram/RCS/vc_main.c,v $
 * $Revision: 1.1 $
 * $Date: 1998/04/15 08:47:29 $
 * $Log: vc_main.c,v $
 * Revision 1.1  1998/04/15  08:47:29  marzetta
 * Initial revision
 *
 *
 */

#define LEFTRIGHTREC0

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/utsname.h>
#include "z_bb.h"
#include "z_time.h"
#include "z_par.h"
#include "z_io.h"		/* for saving the global data */
#include "z_signal.h"
#include "vc_base.h"
#include "vc_io.h"
#include "vc_lb.h"
#include "vc_br.h"
#include "vc_ub.h"
#include "vc_init.h"		/* vc_procsetnum */

#ifdef sunbsd
int getopt(int argc, char **argv, char *optstring);
extern char *optarg;
extern int optind, opterr;
#endif

static vc_NodeP root;
static vc_GlobalP global;
static FILE *logfile;
static boolean printsol;

static void CheckSolution (vc_GlobalP global, vc_NodeP root, vc_NodeP sol)
{
  MINT i;
  
  assert(root->vc_numPt == sol->vc_numPt);
  assert(root->vc_numMarked == 0);
  assert(sol->vc_numAct == 0);
  assert(sol->vc_numEdge == 0);
  for (i=0; i<root->vc_numPt; i++) {
    if (vc_degree(sol)[i] != MARK_DEGREE) {
      MINT *adj;
      MINT other;
      
      adj = vc_adjPt(global, i);
      while (*adj != ADJ_LAST) {
	other = vc_pt(*adj);
	if (vc_degree(sol)[other] != MARK_DEGREE)
	  printf("ERROR: Edge (%ld,%ld) is not marked.\n",
		 (long)i, (long)other);
	  adj++;
      }
    }
  }
} /* CheckSolution */

static void InitRoot(void)
{
  int i;
  
  root->vc_numAct = global->vc_numAct;
  for (i = 0; i < global->vc_numPt; i++)
    vc_degree(root)[i] = vc_gdegree(global)[i];
  /*vc_PrintNode(global, root);*/
}

#ifdef LEFTRIGHT

static void InitLeftRightLB(void)
{
  /* Modifies global->vc_lbProc */

  vc_NodeP np, sol;
  int lbLeft;
  
  np = (vc_NodeP) z_NewCopyNode((z_NodeP) root);
  Restrict(global, np, 1);
  /*vc_PrintNode(global, np);*/
  sol = (vc_NodeP)z_BranchBoundMixedFirst(vc_procsetnum,
					  (z_bb_GlobalDataP) global,
					  (z_NodeP)np, 10000, 1);
  lbLeft = sol->vc_lb;
  printf("Left LB: %ld\n", (long) sol->vc_lb);
  (void) vc_FreeNode(np);
  (void) vc_FreeNode(sol);
  np = (vc_NodeP) z_NewCopyNode((z_NodeP) root);
  Restrict(global, np, 0);
  /*vc_PrintNode(global, np);*/
  sol = (vc_NodeP)z_BranchBoundMixedFirst(vc_procsetnum,
					  (z_bb_GlobalDataP) global,
					  (z_NodeP)np, 10000, 1);
  global->vc_lbRight = sol->vc_lb;
  global->vc_lbLeft = lbLeft;
  printf("Right LB: %ld\n", (long) sol->vc_lb);
  (void) vc_FreeNode(np);
  (void) vc_FreeNode(sol);
} /* InitLeftRightLB */

#endif /* LEFTRIGHT */

static void CountTriangles(void *ref, int a, int b, int c)
{
  (* (int *) ref) ++;
  if (vc_triangle(global)[a] == -1)
    vc_triangle(global)[a] = b;
  /*printf("Triangle:        %3d%3d%3d\n", a, b, c);*/
} /* CountTriangles */

static void InitProblem(char *name)
{
  /* Read the graph from the file name.
     Initialize global and root.
     Initialize global->vc_lbProc.
     Initialize root->vc_lb.
     */
  FILE *infile;
  
  infile = fopen(name, "r");
  if (! infile) {
    perror(name);
    global = NULL;
    return;
  }
  
  global = vc_ReadGraph(infile);
  if (global) {
    global->vc_procsetnum = vc_procsetnum;
    root = vc_NewShortNode(global->vc_numPt, global->vc_numEdge);
    if (root) {
      InitRoot();
      /*vc_PrintNode(global, root);*/
      global->vc_lbProc = 4;
      {
	int numTriangles;
	numTriangles = 0;
        ForAllTriangles(global, root, CountTriangles, &numTriangles);
	printf("Triangles: %d\n", numTriangles);
	if (numTriangles == 0) global->vc_lbProc = 2;
      }
#ifdef LEFTRIGHTREC
      global->vc_lbProc = 6;
#endif
#ifdef LEFTRIGHT
      InitLeftRightLB();
      global->vc_lbProc = 5;
#endif
      printf("Using Lower Bound %ld\n", (long) global->vc_lbProc);
      root->vc_lb = vc_LowerBound(global, root);
    } /* root */
  } /* global */
} /* InitProblem */


static void LogLB(char *name)
{
  InitProblem(name);
  if (global && root) {
    printf("%s\tRoot: (%s)\n", name, vc_ShortNodeString(root));
    fprintf(logfile, "%s\t%4ld (lb%ld)\n", name,
	    (long) root->vc_lb, (long) global->vc_lbProc);
    fflush(logfile);
  } /* global && root */
} /* LogLB */

static void InitLogfile(void)
{
  if (getenv("VCLOG")) {
    logfile = fopen(getenv("VCLOG"), "a+");
    if (! logfile) {
      perror(getenv("VCLOG"));
      logfile = stdout;
    }
  } else
    logfile = stdout;
}

static void Estimate(long numEstimates, long ub)
{
  float nNodes, sNodes, nTime, sTime;
  vc_NodeP upperBound;

  upperBound = (vc_NodeP) z_NewCopyNode((z_NodeP) root);
  upperBound->vc_lb = ub;
  
  z_BranchBoundEstimate(vc_procsetnum, (z_NodeP) global, (z_NodeP) root,
			(z_NodeP) upperBound,
			numEstimates, &nNodes, &sNodes,
			&nTime, &sTime);
  printf("Estimate: %5.1f +- %5.1f nodes in search tree\n", nNodes, sNodes);
  printf("Estimated time: %5.4f +- %5.4fs\n", nTime, sTime);
  
  z_FreeNode((z_NodeP)upperBound);
} /* Estimate */

static void Solve(boolean sequential, char *name)
{
  double time;
  char s[80], hostname[5];
  int queuesize;
  vc_NodeP sol;
  struct utsname utsname;
    
    printf("Root: (%s)\n\n", vc_ShortNodeString(root));
    queuesize = (z_memsize - 2000000) / (root->len + 8);
    /*printf("%ld %ld %ld %ld\n",(z_memsize - 2000000) , (long)(root->len + 8)
	   , (long)queuesize, sizeof(degreeT));*/
    /*queuesize = 80000;*/
				/* q=100000, 300pts(vc10) -> 129 MByte */
				/* q=65000                -> 88.2MByte */
				/* 90 MByte would be sensible */
    time = z_Gtime();
#ifndef SEQUENTIAL
    if (! sequential)
      sol = (vc_NodeP)z_BranchBoundParallel(vc_procsetnum,
					    (z_bb_GlobalDataP) global,
					    (z_NodeP) root, NULL,
					    queuesize, 1);
    else
#endif /* SEQUENTIAL */
      sol = (vc_NodeP)z_BranchBoundMixedFirst(vc_procsetnum,
					     (z_bb_GlobalDataP) global,
					     (z_NodeP)root, NULL,
					      queuesize, 1);
    time = z_Gtime() - time;
    if (uname(&utsname) < 0) perror("uname");
    strncpy(hostname, utsname.nodename, 4);
    hostname[4] = 0;
    sprintf(s, "%-16sProcs:%3d(%-4s) Time:%7.2f (lb%ld) q:%6d",
	    name, z_numprocs, hostname, time,
	    (long)global->vc_lbProc, queuesize);
    fprintf(logfile, "%s\n", s);
    printf("%s\n", s);
    CheckSolution(global, root, sol);
    printf("Solution: (%s)\n", vc_ShortNodeString(sol));
    PrintLbHisto(name, sol->vc_lb);
    if (printsol) vc_PrintNode(global, sol);

} /* Solve */

const char globalFileName[] = "gf"; /* should be read from command line */

static void SolveDF(boolean sequential, char *name, int ub)
{
  double time;
  char s[80], hostname[5];
  vc_NodeP sol;
  struct utsname utsname;
  vc_NodeP upperBound;

  upperBound = (vc_NodeP) z_NewCopyNode((z_NodeP) root);
  upperBound->vc_lb = ub;
  printf("Root: (%s)\n\n", vc_ShortNodeString(root));
  {
    FILE *gf;
      
    gf = fopen(globalFileName, "w");	/* save for restart */
    if (!gf) {
      perror("open/w");
      return;
    }
    z_WriteNode(gf, (z_NodeP) global);
    fclose(gf);
  }
  time = z_Gtime();
  z_StartCheckpointTimer();
  sol = (vc_NodeP)z_BranchBoundDFExp(vc_procsetnum,
				     (z_bb_GlobalDataP) global,
				     (z_NodeP) root, (z_NodeP) upperBound,
				     1, sequential);
  time = z_Gtime() - time;
  if (uname(&utsname) < 0) perror("uname");
  strncpy(hostname, utsname.nodename, 4);
  hostname[4] = 0;
  sprintf(s, "%-16sProcs:%3d(%-4s) Time:%7.2f (lb%ld)",
	  name, z_numprocs, hostname, time,
	  (long)global->vc_lbProc);
  fprintf(logfile, "%s\n", s);
  printf("%s\n", s);
  /*CheckSolution(global, root, sol);*/
  printf("Solution: (%s)\n", vc_ShortNodeString(sol));
  if (printsol) vc_PrintNode(global, sol);
} /* SolveDF */

int Restart(boolean sequential)
{
  FILE *gf;
  vc_GlobalP global;
  vc_NodeP sol;

  printf("***** Restarting from checkpoint\n");
  gf = fopen(globalFileName, "r");
  if (!gf) {
    perror(globalFileName);
    return 9;
  }
  global = (vc_GlobalP) z_ReadNode(gf);
  fclose(gf);
  z_StartCheckpointTimer();
  sol = (vc_NodeP) z_RestartBranchBoundDFExp(vc_procsetnum,
					     (z_bb_GlobalDataP) global,
					     sequential);
  printf("Solution: (%s)\n", vc_ShortNodeString(sol));
  if (printsol) vc_PrintNode(global, sol);
  return 0;
} /* Restart */

static int Usage(const char *name)
{
  fprintf(stderr, "Usage: %s [-1dehr] infile\n", name);
  fprintf(stderr, "       %s -l\n", name);
  fprintf(stderr, "       %s -r\n\n", name);
  fprintf(stderr, "Options: -1         sequential\n");
  fprintf(stderr, "         -d ub      depth-first, parallel, restartable\n");
  fprintf(stderr, "         -e num ub  estimate size of search tree\n");
  fprintf(stderr, "         -h         help (produce this message)\n");
  fprintf(stderr, "         -k nin nout redistribute checkpoint files\n");
  fprintf(stderr, "         -l         ...\n");
  fprintf(stderr, "         -p         print the solution (not just its value)\n");
  fprintf(stderr, "         -r         restart\n");
  return 4;
} /* Usage */

int mmain(int argc, char **argv, char **envp)
{
  char *name;
#ifdef SEQUENTIAL
  boolean sequential=1;
#else
  boolean sequential=0;
#endif /* SEQUENTIAL */
  boolean estimate = 0, depthfirst = 0, restart = 0;
  long numEstimates, ub;
  char c;
  int nin, nout;

  InitLogfile();
  
  while ((c=getopt(argc, argv, "1d:e:chk:pr")) != -1)
    switch(c) {
    case '1':
      sequential = 1;
      break;
    case 'd':
      depthfirst = 1;
      ub = atoi(optarg);
      printf("ub = %ld\n", ub);
      break;
    case 'e':
      estimate = 1;
      numEstimates = atoi(optarg);
      ub = atoi(argv[optind++]);
      printf("numEstimates = %ld, ub = %ld\n", numEstimates, ub);
      break;
    case 'k':
      nin = atoi(optarg);
      nout = atoi(argv[optind++]);
      return z_bb_RedistributeDFExp(vc_procsetnum, nin, nout);
      /*break;*/
    case 'l':
      {
	int i;
	char name[100];
	for (i = 1; i <= 13; i++) {
	  sprintf(name, "vc%d", i);
	  LogLB(name);
	}
	return 0;
      }
      break;
    case 'p':
      printsol = 1;
      break;
    case 'r':
      restart = 1;
      break;
    default:
      return Usage(argv[0]);
    }

  if (restart) {
    if (argc != optind) return Usage(argv[0]);
    return Restart(sequential);
  }
  if (argc != optind+1) return Usage(argv[0]);

  name = argv[optind];
  
  InitProblem(name);
  if (global && root) {
    if (estimate) Estimate(numEstimates, ub);
      else if (depthfirst) SolveDF(sequential, name, ub);
      else Solve(sequential, name);
  } /* global && root */
  return 0;
} /* mmain */

