/* z_mp.c (message passing with termination detection)

   Ambros Marzetta

   http://wwwjn.inf.ethz.ch/ambros/zram.html
   
 * $Source: /home/schilte/marzetta/work/distr/zram/RCS/z_mp.c,v $
 * $Revision: 1.4 $
 * $Date: 1998/03/10 15:02:16 $
 * $Log: z_mp.c,v $
 * Revision 1.4  1998/03/10  15:02:16  marzetta
 * z_Poll() is now a boolean function rather than void.
 *
 * Revision 1.3  1996/04/09  13:24:10  marzetta
 * Added new Install procedures and SEQUENTIAL
 *
 * Revision 1.2  1995/09/14  12:26:23  marzetta
 * *** empty log message ***
 *
 * Revision 1.1  1995/08/23  10:04:12  marzetta
 * Initial revision
 *
 */

#include <stdio.h>
#include <stdlib.h>		/* abort */
#include <string.h>
#include "z_par_intern.h"	/* inbuffer */
#include "z_mp.h"
#include "z_md.h"

#define TRACESEND0
#define TRACERECEIVE0
#define TRACERECEIVE10
#define TRACETOKEN0
#define TRACEIDLE0
#define TRACECALLHANDLER0

#ifndef SEQUENTIAL
/* Message passing with termination detection ---------------------------- */

static boolean mp_inited;	/* set by z_mp_Init */
static boolean mpStarted;	/* set by z_StartMessagePassing / z_Idle */
static boolean cansend = 1;
static boolean inHandler;	/* make handlers / senders non-reentrant */

static struct buffer_header termbuffer = {2234576, 0};
static long *msgCount, *token; /* allocated by z_mp_Init */
static struct buffer *tokenbuffer; /* allocated by z_mp_Init */
static long tokensize;
static boolean tokenHere;	/* the token is on this PE */
				/* modified by z_StartMessagePassing, */
				/* GotToken, z_Idle */
static boolean tokenOtherZero;	/* token here and contains zeroes for */
				/* all other PEs */
				/* modified by z_StartMessagePassing, */
				/* GotToken, z_Isend, z_Ibroadcast */
static boolean tokenAllZero; /* token here and contains zeroes for all PEs */

enum IdleStat {is_terminate, is_sent, is_received};

#define TOKEN 27388
#define isToken (inbuffer->header.msg_tag == TOKEN)
#define isTermtag (inbuffer->header.msg_tag == termbuffer.msg_tag)

#ifdef TRACETOKEN
static void PrintToken(void)
{
  int i;
  assert(tokenHere);
  printf("[%2d]", z_myproc);
  for (i = 0; i < z_numprocs; i++) printf("%4d", token[i]);
  printf("\n");
}
#else
#define PrintToken() (void)0
#endif /* TRACETOKEN */

static void IrecvIfNotAllZero(void)
{
  if (! tokenAllZero) {
    z_md_Irecv();
    receiving = 1;
  }
} /* IrecvIfNotAllZero */

static void z_StartReceive(void)
{
  assert(z_myproc != -1);
  assert(! receiving);
  assert(tokenAllZero == (tokenOtherZero && token[z_myproc] == 0));
  IrecvIfNotAllZero();
  inpos = 0;
} /* z_StartReceive */

static void GotToken(void)
{
  int i;

  memcpy(token, inbuffer->b, z_numprocs * sizeof(*token));
  tokenHere = 1;
  for_all_nodes(i) {
    token[i] += msgCount[i];
    msgCount[i] = 0;
  }
  tokenOtherZero = 1;
  for_all_nodes(i)
    if (i != z_myproc && token[i]) tokenOtherZero = 0;
  tokenAllZero = tokenOtherZero && token[z_myproc] == 0;
} /* GotToken */

static boolean ReceivedMessage(void)
{
  /* Pre: inbuffer contains a token or a user message. ! receiving
     Return 1 if a user message was received
     Post: 0 -> a token was received and processed
                receiving | tokenAllZero
	   1 -> a user message was received
	        the user message has been counted
	        ! receiving
     */
  if (isToken) {
    GotToken();
    IrecvIfNotAllZero();
    assert(inpos == 0);
    return 0;
    /* receive in progress */
  } else {			/* received user message */
    if (tokenHere) {
      token[z_myproc] --;
      tokenAllZero = tokenOtherZero && token[z_myproc] == 0;
    } else
      msgCount[z_myproc] --;
    assert(! receiving);
    return 1;
  }
} /* ReceivedMessage */

static boolean z_GotMessage(void)
{
  /* Pre: there is a receive pending */
  assert(z_myproc != -1);
  if (tokenAllZero) return 0; /* I'm the only one working. */
  assert(receiving);
  if (z_md_TestRecv()) {
    receiving = 0;
    return ReceivedMessage();
  }
  return 0;
} /* z_GotMessage */

void z_StartMessagePassing(void)
{
  /* Start message passing and the termination detection algorithm.
   */
  assert(z_myproc != -1);
  assert(mp_inited);
  assert(! inHandler);
  assert(! mpStarted);
  mpStarted = 1;
  /* Initialize the termination detection algorithm. */
  {
    int i;
    for_all_nodes(i)
      msgCount[i] = 0;		/* I have not yet sent messages */
    if (! z_md_doesTD) {	/* do termination detection here */
      tokenHere = z_myproc == ROOT_PE;
      tokenOtherZero = tokenAllZero = (z_numprocs == 1);
      if (tokenHere) {
	for_all_nodes(i)
	  token[i] = 1;		/* FE has sent one message to each PE */
	token[z_myproc] = 0;	/* I have received my message */
      } else
	msgCount[z_myproc] = -1;	/* I have received my message */
    } else
      tokenHere = tokenOtherZero = tokenAllZero = 0;
  }
  z_StartReceive();
} /* z_StartMessagePassing */

boolean z_TestSend(void)
{
  /* (PE only) Is it possible to overwrite outbuffer? */
  assert(mpStarted);
  if (cansend) return 1;
  cansend = z_md_TestSend();
  return cansend;
} /* z_TestSend */

void z_Isend(int dest, char *comment)
{
  /* Send outbuffer to dest*/
  assert(mpStarted);
  assert(cansend);
  assert(z_myproc != dest);
  assert(0 <= dest);
  assert(dest < z_numprocs);
#ifdef TRACESEND
  printf("[%2d]#SendTo %d\n", z_myproc, dest);
#endif
  z_md_Isend(outbuffer,
	      outpos * sizeof(unitT) + sizeof(struct buffer_header),
	      dest, comment);
  if (tokenHere) {
    token[dest] ++;
    tokenOtherZero = 0;
    if (tokenAllZero) {
      tokenAllZero = 0;
      z_md_Irecv();
      receiving = 1;
    }
  } else
    msgCount[dest] ++;
  cansend = 0;
  outpos = 0;
} /* z_Isend */

void z_Ibroadcast(char *comment)
{
  /* Broadcast outbuffer */
  int i;
  assert(z_myproc != -1);
  assert(mpStarted);
  assert(cansend);
#ifdef TRACESEND
  printf("[%2d]#Broadcast\n", z_myproc);
#endif
  z_md_Ibroadcast(outbuffer,
		 outpos * sizeof(unitT) + sizeof(struct buffer_header),
		 comment);
  if (tokenHere) {
    for_all_nodes(i)
      if (i != z_myproc) token[i] ++;
    tokenOtherZero = 0;
    if (tokenAllZero) {
      tokenAllZero = 0;
      z_md_Irecv();
      receiving = 1;
    }
  } else
    for_all_nodes(i)
      if (i != z_myproc) msgCount[i] ++;
  cansend = 0;
  outpos = 0;
} /* z_Ibroadcast */

#define SQSIZE 10
static z_VoidProc sendQueue[SQSIZE];
static int sq_in, sq_out;

void z_QueueSender(z_VoidProc sender)
/* Insert sender into the send queue. As soon as sending becomes possible,
   the first sender from the queue will be called.
   The sender may call z_Isend, z_Ibroadcast and z_QueueSender.
   */
{
  assert(mpStarted);
  assert(sender);
  sendQueue[sq_in] = sender;
  sq_in = (sq_in + 1) % SQSIZE;
  assert(sq_out != sq_in);
} /* z_QueueSender */

static enum IdleStat z_Idle(boolean sureReceive)
/* The processing element is idle.
   There is a receive pending.
   Wait for the receive operation started by z_StartReceive to complete.
   Post: A receive is pending only if the return value is is_sent.

   Invariant of termination detection algorithm:
   tokenHere == (the token is on this PE)
   token + sum of all msgCount == number of messages underway to a PE
   tokenHere -> (msgCount == 0)
   tokenHere -> (tokenOtherZero == (all other elements of the token == 0))
   tokenAllZero == (tokenHere && all elements of the token == 0)
   tokenAllZero == (! receiving)
   ! (ex negatives Token und kein positives)
   */
{
  boolean user, terminate;
  
  assert(mpStarted);
  if (sq_in != sq_out) {  /* busy wait until either send or recv possible */
    while (1) {
      if (z_TestSend()) {
	inHandler = 1;
	sendQueue[sq_out]();	/* call the sender */
	inHandler = 0;
	sq_out = (sq_out + 1) % SQSIZE;
	return is_sent;
      }
      if (z_GotMessage())
	return is_received;
    }
  }
  /* There is no sender in the queue */
  if (z_md_doesTD) {
    receiving = 0;
    terminate = z_md_WaitRecv();
    mpStarted = ! terminate;
    return terminate ? is_terminate : is_received;
  }
  assert((receiving && !tokenAllZero) || (!receiving && tokenAllZero));
  assert(!sureReceive || receiving);
#ifdef TRACEIDLE
  printf("[%2d]#Idle start %c\n", z_myproc, tokenHere ? 'T' : '.');
#endif
  while (1) {
    assert((receiving && !tokenAllZero) || (!receiving && tokenAllZero));
    if (tokenHere) {
      int i;
      PrintToken();
      for_all_nodes(i) assert(msgCount[i] == 0);
    }
    if (!sureReceive &&
	tokenHere &&
	token[z_myproc] <= 0) { /* Test token and send */
      int i;

      i = z_numprocs - 1;
      while (i >= 0 && token[i] <= 0) i --;
      if (i < 0) {		/* token contains only zeroes */
	/*if (! tokenAllZero) {*/
	/*if (tokenOtherZero) printf("tokenOtherZero %ld\n", token[z_myproc]);*/
	/*}*/
	assert(tokenAllZero);
	assert(! receiving);
#ifdef TRACETOKEN
	printf("[%2d] Terminate Job!\n", z_myproc);
#endif
	while (!z_TestSend()) ;	/* busy waiting */
	z_md_Ibroadcast((struct buffer *)& termbuffer,
		       sizeof(termbuffer), "terminate");
	cansend = 0;
	mpStarted = 0;
	return is_terminate;
      }
      /* Send token to processor i */
      while (!z_TestSend()) ;	/* busy waiting */
      z_md_Isend(tokenbuffer, tokensize, i, "token");
      cansend = 0;
#ifdef TRACETOKEN
      printf("[%2d]T Token -> %d\n", z_myproc, i);
#endif
      tokenHere = tokenOtherZero = 0;
    }
    assert(sureReceive || ! tokenHere || token[z_myproc] > 0);
				/* token is not here */
    assert(receiving);
    (void) z_md_WaitRecv();
    receiving = 0;		/* no receive in progress */
    if (isTermtag) {		/* other PE detected termination */
      mpStarted = 0;
      return is_terminate;	/* no receive in progress */
    }
    user = ReceivedMessage();	/* token or user message */
    if (user) return is_received;
  } /* while */
} /* z_Idle */

/* Message handlers -------------------------------------------------- */

#define MESSAGEBASE 12000
#define MAXMESSAGE 100
static int numMessage;
static z_VoidProc messageHandler[MAXMESSAGE];
static const char *messageName[MAXMESSAGE];

z_tagT z__InstallMessage(z_VoidProc handler, const char *name)
/* (PE only) Install a handler for incoming messages with some tag */
{
  messageHandler[numMessage] = handler;
  messageName[numMessage] = name;
  return MESSAGEBASE + numMessage ++;
} /* z_InstallMessage */

static void z_CallHandler(void)
/* Pre: z_GotMessage or z_Idle have returned true.
   Post: z_StartReceive is not called */
{
  int msg_tag;

  assert(! inHandler);
  msg_tag = z_GetTagBuffer();
#ifndef NDEBUG
  if(msg_tag < MESSAGEBASE || msg_tag >= MESSAGEBASE + numMessage) {
    printf("*%2d*z_CallHandler: Illegal message %ld %ld\r\n",
	   z_myproc, (long)msg_tag, (long)z_GetMINTBuffer());
    z_md_abort();
  }
#endif /* ! NDEBUG */
#ifdef TRACECALLHANDLER
  printf("[%2d]z_CallHandler: %s\n", z_myproc,
	 messageName[msg_tag - MESSAGEBASE]);
#endif
  inHandler = 1;
  messageHandler[msg_tag - MESSAGEBASE]();
  inHandler = 0;
} /* z_CallHandler */

boolean z_Poll(void)
/* (PE only) Check incoming messages and call their handler, if any.
   z_StartReceive must have been called. */
{
  boolean done;
  
  assert(mpStarted);
  assert(! inHandler);
  if (sq_in != sq_out && z_TestSend()) {
    inHandler = 1;
    sendQueue[sq_out]();
    inHandler = 0;
    sq_out = (sq_out + 1) % SQSIZE;
    done = 1;
  } else done = 0;
  while (z_GotMessage()) {
    z_CallHandler();
    z_StartReceive();
    done = 1;
  }
  return done;
} /* z_Poll */

boolean z_IdleWithHandler(boolean sureReceive)
{
  enum IdleStat is;
  
  assert(! inHandler);
  /* receive pending */
  is = z_Idle(sureReceive);
  /* no receive pending */

  switch (is) {
  case is_terminate:
    return 1;
  case is_sent:
    return 0;
  case is_received:
    z_CallHandler();
    z_StartReceive();
    return 0;
  default:
    return 0; /* compiler warning: control reaches end of non-void function */
  }
} /* z_IdleWithHandler */


/* Initialization --------------------------------------------------- */

void z_mp_Init()
{
  if (! mp_inited) {
    mp_inited = 1;
    tokensize = z_numprocs * sizeof(long) + sizeof(struct buffer_header);
    msgCount = (long *) malloc(z_numprocs * sizeof(long));
    tokenbuffer = (struct buffer *) malloc(tokensize);
    token = (long *)tokenbuffer->b;
    tokenbuffer->header.msg_tag = TOKEN;
  }
} /* z_mp_Init */
#endif /* SEQUENTIAL */
